"""Definition of ListQuerySet."""

from collections import OrderedDict
from dataclasses import dataclass
from typing import Any


class QuerySetQueryStub:
    """A stub to implement the query property."""

    def __init__(self) -> None:
        self.select_related = None
        self.order_by = []


@dataclass
class _FilterObj:
    """A filter object."""

    field: str
    lookup: str
    value: str

    def is_matching(self, obj: Any) -> bool:
        """Return True, when obj is matching this filter, else False."""
        obj_value = str(getattr(obj, self.field))
        if self.lookup == "exact":
            return self.value == obj_value

        if self.lookup == "contains":
            return self.value in obj_value

        if self.lookup == "icontains":
            return self.value.lower() in obj_value.lower()

        raise NotImplementedError(f"Unknown lookup: {self.lookup}")

    @classmethod
    def create(cls, key: str, value: Any) -> "_FilterObj":
        """Create new filter from key/value pair."""
        parts = key.split("__")

        if len(parts) == 1:
            field = key
            lookup = "exact"

        elif len(parts) == 2:
            field, lookup = parts

        else:
            raise ValueError(f"Invalid key in filter: {key}")

        return cls(field=field, lookup=lookup, value=str(value))


class ListAsQuerySet(list):
    """Masquerade a list as QuerySet.

    Args:
        - model: Django model (mandatory)
        - is_distinct: Whether result should be distinct (optional)
    """

    def __init__(self, *args, model, **kwargs):
        self.model = model
        self.is_distinct = kwargs.pop("is_distinct", False)
        self.query = QuerySetQueryStub()

        super().__init__(*args, **kwargs)
        self._id_mapper = {str(obj.id): n for n, obj in enumerate(self)}
        self._list_size = len(self)

    def all(self) -> "ListAsQuerySet":
        """:private:"""
        if self.is_distinct:
            return self._make_list_distinct(self)
        return self

    def count(self) -> int:
        """:private:"""
        return self._list_size

    def _clone(self):
        return self._make_clone()

    def _make_clone(self, new_list: list = None) -> "ListAsQuerySet":
        if new_list is None:
            new_list = self
        obj = type(self)(list(new_list), model=self.model, is_distinct=self.is_distinct)
        return obj

    def create(self, *args, **kwargs) -> Any:
        """:private:"""
        raise NotImplementedError("ListAsQuerySet can not write to the database.")

    def distinct(self) -> "ListAsQuerySet":
        """:private:"""
        self.is_distinct = True
        return self

    def first(self) -> Any:
        """:private:"""
        try:
            return self[0]
        except IndexError:
            return None

    def filter(self, *args, **kwargs) -> "ListAsQuerySet":
        """:private:

        QuerySet filter.
        Supports args or kwargs, but not both.
        Supports equal clauses only.
        """
        print(args)
        if args:
            new_list = []
            for item in self:
                matches = True
                for q_obj in args:
                    filters = dict(q_obj.children)
                    matches &= all(
                        str(getattr(item, key, "")) == str(val)
                        for key, val in filters.items()
                    )
                if matches:
                    new_list.append(item)

            return self._make_clone(new_list=new_list)

        if kwargs:
            filter_objs = [
                _FilterObj.create(key, value) for key, value in kwargs.items()
            ]

            new_list = [
                item
                for item in self
                if all(filter_obj.is_matching(item) for filter_obj in filter_objs)
            ]

            return self._make_clone(new_list=new_list)

        return self

    def get(self, *args, **kwargs) -> Any:
        """:private:"""
        try:
            return self[self._id_mapper[str(kwargs["id"])]]
        except KeyError:
            raise self.model.DoesNotExist from None

    def last(self) -> Any:
        """:private:"""
        try:
            return self[-1]
        except IndexError:
            return None

    def none(self) -> "ListAsQuerySet":
        """:private:"""
        return self._make_clone(new_list=[])

    def order_by(self, *args, **kwargs):
        """:private:"""
        if not args and not kwargs:
            return self

        if kwargs:
            raise NotImplementedError("order with kw args not supported.")

        new_list = list(self)
        for prop in reversed(args):
            if prop[0:1] == "-":
                reverse = True
                prop_2 = prop[1:]
            else:
                reverse = False
                prop_2 = prop

            # pylint: disable = cell-var-from-loop
            new_list.sort(key=lambda d: getattr(d, prop_2), reverse=reverse)

        return self._make_clone(new_list=new_list)

    def values(self, *args):
        """:private:"""
        result = list(self._values(*args))
        if self.is_distinct:
            return list(
                OrderedDict((frozenset(item.items()), item) for item in result).values()
            )
        return result

    def _values(self, *args):
        result = (
            {k: v for k, v in obj.__dict__.items() if not args or k in args}
            for obj in self
        )
        return result

    def values_list(self, *args, **kwargs):
        """:private:"""
        items = (tuple(obj.values()) for obj in self._values(*args))
        if kwargs.get("flat"):
            if len(args) > 1:
                raise TypeError(
                    "'flat' is not valid when values_list is called with "
                    "more than one field."
                )
            items = (obj[0] for obj in items)
            if self.is_distinct:
                return self._make_list_distinct(items)

        result = list(items)
        return result

    @staticmethod
    def _make_list_distinct(items):
        return list(dict.fromkeys(items))



"""Deal with app names of tasks."""

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from taskmonitor import __title__
from taskmonitor.app_settings import (
    TASKMONITOR_APP_NAME_MAPPING_CONFIG,
    TASKMONITOR_APP_NAME_MAPPING_DEFAULTS,
)

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


def from_task_name(task_name: str) -> str:
    """Return the app name from a typical task name, if possible.
    Otherwise return an empty string.
    """
    app_name = _extract_app_name(task_name)
    app_name = _map_app_names(app_name)
    return app_name


def _extract_app_name(task_name: str):
    """Return the app name extracted from a task name or an empty string."""
    parts = task_name.split(".")
    try:
        idx = parts.index("tasks")
    except ValueError:
        if len(parts) == 2:
            return parts[0]
        return ""

    app_name = parts[idx - 1] if idx > 0 else ""
    return app_name


def _map_app_names(app_name: str):
    """Return the mapped app name (if any) or return itself."""
    new_app_name = _app_names_mapping.get(app_name, app_name)
    return new_app_name


def _consolidate_app_names_mapping():
    """Return map of each section and their time until an update is stale."""
    mapping = {
        alternate_name: name
        for name, alternate_names in TASKMONITOR_APP_NAME_MAPPING_DEFAULTS.items()
        for alternate_name in alternate_names
    }
    for name, alternate_names in TASKMONITOR_APP_NAME_MAPPING_CONFIG.items():
        try:
            for alternate_name in alternate_names:
                mapping[alternate_name] = name
        except TypeError:
            logger.warning(
                "Mapping for %s is invalid and will be ignored. "
                "Please correct or remove it.",
                name,
            )

    return mapping


_app_names_mapping = _consolidate_app_names_mapping()



"""Settings for Task Monitor."""

from app_utils.app_settings import clean_setting

TASKMONITOR_ENABLED = clean_setting("TASKMONITOR_ENABLED", True)
"""Global switch to enable/disable task monitor."""

TASKMONITOR_DATA_MAX_AGE = clean_setting("TASKMONITOR_DATA_MAX_AGE", 24)
"""Max age of logged tasks in hours. Older logs be deleted automatically."""

TASKMONITOR_DELETE_STALE_BATCH_SIZE = clean_setting(
    "TASKMONITOR_DELETE_STALE_BATCH_SIZE", 5_000
)
"""Size of task logs deleted together in one batch."""

TASKMONITOR_HOUSEKEEPING_FREQUENCY = clean_setting(
    "TASKMONITOR_HOUSEKEEPING_FREQUENCY", 30
)
"""Frequency of house keeping runs in minutes."""

TASKMONITOR_REPORTS_MAX_AGE = clean_setting("TASKMONITOR_REPORTS_MAX_AGE", 60)
"""Max age of cached reports in minutes."""


TASKMONITOR_QUEUED_TASKS_CACHE_TIMEOUT = clean_setting(
    "TASKMONITOR_QUEUED_TASKS_CACHE_TIMEOUT", 10
)
"""Timeout for caching queued tasks in seconds. 0 will deactivate the cache."""

TASKMONITOR_QUEUED_TASKS_ADMIN_LIMIT = clean_setting(
    "TASKMONITOR_QUEUED_TASKS_ADMIN_LIMIT", 100_000
)
"""The admin page will stop showing the list of queued tasks above this limit
to protect against crashing caused by too high memory consumption.
"""

TASKMONITOR_REPORTS_MAX_TOP = clean_setting("TASKMONITOR_REPORTS_MAX_TOP", 20)
"""Max items to show in the top reports. e.g. 10 will shop the top ten items."""

TASKMONITOR_TRUNCATE_NESTED_DATA = clean_setting(
    "TASKMONITOR_TRUNCATE_NESTED_DATA", True
)
"""Whether deeply nested task params and results are truncated."""

TASKMONITOR_APP_NAME_MAPPING_CONFIG = clean_setting(
    "TASKMONITOR_APP_NAME_MAPPING_CONFIG", {}
)
"""Ability to map tasks to the same app name.
Map must be a dictionary with string keys and list of strings as value.
All app names in the list will be replaced by it's key.
"""

TASKMONITOR_APP_NAME_MAPPING_DEFAULTS = {"standingsrequests": ["standings_requests"]}
"""Default app mappings.

Can be overridden or extended through TASKMONITOR_APP_NAME_MAPPING_CONFIG.
"""



from unittest.mock import patch

from django.test import TestCase

from taskmonitor.core import app_names

MODULE_PATH = "taskmonitor.core.app_names"


class TestExtractAppName(TestCase):
    def test_can_extract_from_normal_task_name(self):
        # when
        result = app_names.from_task_name("alpha.tasks.task_name")
        # then
        self.assertEqual(result, "alpha")

    def test_can_extract_from_long_task_name(self):
        # when
        result = app_names.from_task_name("omega.alpha.tasks.task_name")
        # then
        self.assertEqual(result, "alpha")

    def test_can_extract_from_extra_long_task_name(self):
        # when
        result = app_names.from_task_name("echo.omega.alpha.tasks.task_name")
        # then
        self.assertEqual(result, "alpha")

    def test_can_extract_from_custom_task_name(self):
        # when
        result = app_names.from_task_name("alpha.task_name")
        # then
        self.assertEqual(result, "alpha")

    def test_should_return_empty_string_if_no_match_1(self):
        # when
        result = app_names.from_task_name("dummy")
        # then
        self.assertEqual(result, "")

    def test_should_return_empty_string_if_no_match_2(self):
        # when
        result = app_names.from_task_name("tasks.dummy")
        # then
        self.assertEqual(result, "")


class TestAppNameMapping(TestCase):
    @patch(MODULE_PATH + "._app_names_mapping", {"alpha": "bravo"})
    def test_should_map_to_alternate_name(self):
        # when
        result = app_names.from_task_name("alpha.tasks.task_name")
        # then
        self.assertEqual(result, "bravo")

    @patch(MODULE_PATH + "._app_names_mapping", {"charlie": "bravo"})
    def test_should_not_map(self):
        # when
        result = app_names.from_task_name("alpha.tasks.task_name")
        # then
        self.assertEqual(result, "alpha")


class TestConsolidateApp(TestCase):
    @patch(MODULE_PATH + ".TASKMONITOR_APP_NAME_MAPPING_CONFIG", {"alpha": ["alpha_3"]})
    @patch(
        MODULE_PATH + ".TASKMONITOR_APP_NAME_MAPPING_DEFAULTS", {"alpha": ["alpha_2"]}
    )
    def test_should_return_mapping(self):
        # when
        result = app_names._consolidate_app_names_mapping()
        # then
        expected = {"alpha_2": "alpha", "alpha_3": "alpha"}
        self.assertEqual(result, expected)



from django.db.models import Q
from django.test import TestCase

from taskmonitor.core.list_queryset import ListAsQuerySet
from taskmonitor.models import QueuedTask
from taskmonitor.tests.factories import QueuedTaskFactory


class TestOther(TestCase):
    def test_count(self):
        # given
        data = [QueuedTaskFactory.build(), QueuedTaskFactory.build()]
        qs = ListAsQuerySet(data, model=QueuedTask)
        # when/then
        self.assertEqual(qs.count(), 2)

    def test_none(self):
        # given
        data = [
            QueuedTaskFactory.build(),
            QueuedTaskFactory.build(),
            QueuedTaskFactory.build(),
        ]
        qs = ListAsQuerySet(data, model=QueuedTask)
        # when
        result = qs.none()
        # then
        self.assertEqual(result, [])

    def test_get_should_return_obj(self):
        # given
        t1 = QueuedTaskFactory.build(id=1)
        t2 = QueuedTaskFactory.build(id=2)
        t3 = QueuedTaskFactory.build(id=3)
        data = [t1, t2, t3]
        qs = ListAsQuerySet(data, model=QueuedTask)
        # when
        obj = qs.get(id=3)
        # then
        self.assertEqual(obj, t3)

    def test_get_should_raise_error_when_obj_not_found(self):
        # given
        t1 = QueuedTaskFactory.build(id=1)
        t2 = QueuedTaskFactory.build(id=2)
        t3 = QueuedTaskFactory.build(id=3)
        data = [t1, t2, t3]
        qs = ListAsQuerySet(data, model=QueuedTask)
        # when/then
        with self.assertRaises(QueuedTask.DoesNotExist):
            qs.get(id=4)

    def test_first_should_return_obj_when_it_exists(self):
        # given
        t1 = QueuedTaskFactory.build(id=1)
        t2 = QueuedTaskFactory.build(id=2)
        t3 = QueuedTaskFactory.build(id=3)
        data = [t1, t2, t3]
        qs = ListAsQuerySet(data, model=QueuedTask)
        # when
        obj = qs.first()
        # then
        self.assertEqual(obj, t1)

    def test_first_should_return_none_when_obj_not_exists(self):
        # given
        data = []
        qs = ListAsQuerySet(data, model=QueuedTask)
        # when
        obj = qs.first()
        # then
        self.assertIsNone(obj)

    def test_last_should_return_last_element_when_it_exists(self):
        # given
        t1 = QueuedTaskFactory.build(id=1)
        t2 = QueuedTaskFactory.build(id=2)
        t3 = QueuedTaskFactory.build(id=3)
        data = [t1, t2, t3]
        qs = ListAsQuerySet(data, model=QueuedTask)
        # when
        obj = qs.last()
        # then
        self.assertEqual(obj, t3)

    def test_last_should_return_none_when_obj_not_exists(self):
        # given
        data = []
        qs = ListAsQuerySet(data, model=QueuedTask)
        # when
        obj = qs.last()
        # then
        self.assertIsNone(obj)

    def test_clone(self):
        # given
        t1 = QueuedTaskFactory.build(app_name="alpha", name="one")
        t2 = QueuedTaskFactory.build(app_name="alpha", name="two")
        t3 = QueuedTaskFactory.build(app_name="alpha", name="three")
        data = [t1, t2, t3]
        qs = ListAsQuerySet(data, model=QueuedTask)

        # when
        result = qs._clone()

        # then
        self.assertEqual(result, data)
        self.assertIsNot(result, data)

    def test_should_raise_error_when_trying_to_create(self):
        # when/then
        with self.assertRaises(NotImplementedError):
            QueuedTaskFactory.create()


class TestAll(TestCase):
    def test_should_return_all_objs(self):
        # given
        t1 = QueuedTaskFactory.build()
        t2 = QueuedTaskFactory.build()
        data = [t1, t2]
        qs = ListAsQuerySet(data, model=QueuedTask)
        # when
        result = qs.all()
        # then
        self.assertEqual(result, data)

    def test_should_return_all_objs_with_distinct(self):
        # given
        t1 = QueuedTaskFactory.build()
        data = [t1, t1]
        qs = ListAsQuerySet(data, model=QueuedTask)
        # when
        result = qs.distinct().all()
        # then
        self.assertEqual(result, [t1])


class TestFilter(TestCase):
    def test_filter_no_params(self):
        # given
        t1 = QueuedTaskFactory.build(app_name="alpha", name="one")
        t2 = QueuedTaskFactory.build(app_name="alpha", name="two")
        t3 = QueuedTaskFactory.build(app_name="alpha", name="three")
        data = [t1, t2, t3]
        qs = ListAsQuerySet(data, model=QueuedTask)
        # when
        result = qs.filter()
        self.assertEqual(result, data)

    def test_filter_single_kwargs(self):
        # given
        t1 = QueuedTaskFactory.build(app_name="alpha", name="one")
        t2 = QueuedTaskFactory.build(app_name="alpha", name="two")
        t3 = QueuedTaskFactory.build(app_name="alpha", name="three")
        data = [t1, t2, t3]
        qs = ListAsQuerySet(data, model=QueuedTask)
        # when
        result = qs.filter(name="two")
        self.assertEqual(result, [t2])

    def test_filter_multiple_kwargs(self):
        # given
        t1 = QueuedTaskFactory.build(app_name="alpha", name="1", priority=2)
        t2 = QueuedTaskFactory.build(app_name="alpha", name="2", priority=3)
        t3 = QueuedTaskFactory.build(app_name="alpha", name="3", priority=3)
        data = [t1, t2, t3]
        qs = ListAsQuerySet(data, model=QueuedTask)
        # when
        result = qs.filter(app_name="alpha", priority=3)
        self.assertEqual(result, [t2, t3])

    def test_filter_single_args(self):
        # given
        t1 = QueuedTaskFactory.build(app_name="alpha", name="one")
        t2 = QueuedTaskFactory.build(app_name="alpha", name="two")
        t3 = QueuedTaskFactory.build(app_name="alpha", name="three")
        data = [t1, t2, t3]
        qs = ListAsQuerySet(data, model=QueuedTask)
        # when
        result = qs.filter(Q(name="two"))
        # then
        self.assertEqual(result, [t2])

    def test_filter_multiple_args(self):
        # given
        t1 = QueuedTaskFactory.build(app_name="alpha", name="one")
        t2 = QueuedTaskFactory.build(app_name="alpha", name="two")
        t3 = QueuedTaskFactory.build(app_name="bravo", name="two")
        data = [t1, t2, t3]
        qs = ListAsQuerySet(data, model=QueuedTask)
        # when
        result = qs.filter(Q(name="two"), Q(app_name="alpha"))
        # then
        self.assertEqual(result, [t2])

    def test_filter_args_with_values(self):
        # given
        t1 = QueuedTaskFactory.build(app_name="alpha", priority=1)
        t2 = QueuedTaskFactory.build(app_name="alpha", priority=3)
        data = [t1, t2]
        qs = ListAsQuerySet(data, model=QueuedTask)
        # when
        result = qs.filter(Q(priority="3"))
        # then
        self.assertEqual(result, [t2])

    def test_filter_kwargs_with_contains(self):
        # given
        t1 = QueuedTaskFactory.build(app_name="alpha", name="joe_1")
        t2 = QueuedTaskFactory.build(app_name="alpha", name="joe_2")
        t3 = QueuedTaskFactory.build(app_name="alpha", name="Joe")
        data = [t1, t2, t3]
        qs = ListAsQuerySet(data, model=QueuedTask)
        # when
        result = qs.filter(name__contains="joe")
        self.assertEqual(result, [t1, t2])

    def test_filter_kwargs_with_icontains(self):
        # given
        t1 = QueuedTaskFactory.build(app_name="alpha", name="joe_1")
        t2 = QueuedTaskFactory.build(app_name="alpha", name="Joe_2")
        t3 = QueuedTaskFactory.build(app_name="alpha", name="bob")
        data = [t1, t2, t3]
        qs = ListAsQuerySet(data, model=QueuedTask)
        # when
        result = qs.filter(name__icontains="joe")
        self.assertEqual(result, [t1, t2])

    def test_should_raise_error_when_lookup_is_unknown(self):
        # given
        t1 = QueuedTaskFactory.build(app_name="alpha", name="joe_1")
        data = [t1]
        qs = ListAsQuerySet(data, model=QueuedTask)
        # when/then
        with self.assertRaises(NotImplementedError):
            qs.filter(name__unknown="joe")

    def test_should_raise_error_when_filter_is_invalid(self):
        # given
        t1 = QueuedTaskFactory.build(app_name="alpha", name="joe_1")
        data = [t1]
        qs = ListAsQuerySet(data, model=QueuedTask)
        # when/then
        with self.assertRaises(ValueError):
            qs.filter(name__exact__illegal="joe")

    def test_filter_should_allow_chaining(self):
        # given
        t1 = QueuedTaskFactory.build(app_name="charlie")
        t2 = QueuedTaskFactory.build(app_name="alpha")
        t3 = QueuedTaskFactory.build(app_name="alpha")
        data = [t1, t2, t3]
        qs = ListAsQuerySet(data, model=QueuedTask)

        # when
        result = qs.filter(app_name="alpha").count()

        # then
        self.assertEqual(result, 2)


class TestOrderBy(TestCase):
    def test_should_order_by_field_ascending(self):
        # given
        t1 = QueuedTaskFactory.build(name="charlie")
        t2 = QueuedTaskFactory.build(name="alpha")
        t3 = QueuedTaskFactory.build(name="bravo")
        data = [t1, t2, t3]
        qs = ListAsQuerySet(data, model=QueuedTask)

        # when
        result = qs.order_by("name")

        # then
        self.assertIsInstance(result, ListAsQuerySet)
        self.assertEqual(result, [t2, t3, t1])

    def test_should_order_by_field_descending(self):
        # given
        t1 = QueuedTaskFactory.build(name="charlie")
        t2 = QueuedTaskFactory.build(name="alpha")
        t3 = QueuedTaskFactory.build(name="bravo")
        data = [t1, t2, t3]
        qs = ListAsQuerySet(data, model=QueuedTask)

        # when
        result = qs.order_by("-name")

        # then
        self.assertIsInstance(result, ListAsQuerySet)
        self.assertEqual(result, [t1, t3, t2])

    def test_order_should_return_self_unchanged_when_no_params_given(self):
        # given
        t1 = QueuedTaskFactory.build(name="charlie")
        t2 = QueuedTaskFactory.build(name="alpha")
        t3 = QueuedTaskFactory.build(name="bravo")
        data = [t1, t2, t3]
        qs = ListAsQuerySet(data, model=QueuedTask)

        # when
        result = qs.order_by()

        # then
        self.assertIsInstance(result, ListAsQuerySet)
        self.assertEqual(result, qs)

    def test_should_order_by_multiple_fields_ascending(self):
        # given
        t1 = QueuedTaskFactory.build(name="charlie", priority=1)
        t2 = QueuedTaskFactory.build(name="alpha", priority=2)
        t3 = QueuedTaskFactory.build(name="alpha", priority=1)
        data = [t1, t2, t3]
        qs = ListAsQuerySet(data, model=QueuedTask)

        # when
        result = qs.order_by("name", "priority")

        # then
        self.assertIsInstance(result, ListAsQuerySet)
        self.assertEqual(result, [t3, t2, t1])

    def test_should_order_by_multiple_fields_mixed(self):
        # given
        t1 = QueuedTaskFactory.build(name="charlie", priority=1)
        t2 = QueuedTaskFactory.build(name="alpha", priority=2)
        t3 = QueuedTaskFactory.build(name="alpha", priority=1)
        data = [t1, t2, t3]
        qs = ListAsQuerySet(data, model=QueuedTask)

        # when
        result = qs.order_by("-name", "priority")

        # then
        self.assertIsInstance(result, ListAsQuerySet)
        self.assertEqual(result, [t1, t3, t2])

    def test_should_keep_distinct(self):
        # given
        t1 = QueuedTaskFactory.build(name="charlie", priority=1)
        t2 = QueuedTaskFactory.build(name="alpha", priority=2)
        t3 = QueuedTaskFactory.build(name="alpha", priority=1)
        data = [t1, t2, t3]
        qs = ListAsQuerySet(data, model=QueuedTask)

        # when
        result = qs.distinct().order_by("name").values_list("name", flat=True)

        # then
        self.assertEqual(result, ["alpha", "charlie"])


class TestValues(TestCase):
    def test_should_return_data(self):
        # given
        t1 = QueuedTaskFactory.build(name="charlie", priority=1)
        t2 = QueuedTaskFactory.build(name="alpha", priority=2)
        data = [t1, t2]
        qs = ListAsQuerySet(data, model=QueuedTask)

        # when
        result = qs.values("name", "priority")

        # then
        self.assertEqual(
            result,
            [{"name": "charlie", "priority": 1}, {"name": "alpha", "priority": 2}],
        )

    def test_should_return_data_with_distinct(self):
        # given
        t1 = QueuedTaskFactory.build(name="charlie")
        t2 = QueuedTaskFactory.build(name="alpha")
        t3 = QueuedTaskFactory.build(name="alpha")
        data = [t1, t2, t3]
        qs = ListAsQuerySet(data, model=QueuedTask)

        # when
        result = qs.distinct().values("name")

        # then
        self.assertEqual(result, [{"name": "charlie"}, {"name": "alpha"}])


class TestValuesList(TestCase):
    def test_should_return_as_tuples(self):
        # given
        t1 = QueuedTaskFactory.build(name="charlie", priority=1)
        t2 = QueuedTaskFactory.build(name="alpha", priority=2)
        data = [t1, t2]
        qs = ListAsQuerySet(data, model=QueuedTask)

        # when
        result = qs.values_list("name", "priority")

        # then
        self.assertEqual(result, [("charlie", 1), ("alpha", 2)])

    def test_should_return_flat(self):
        # given
        t1 = QueuedTaskFactory.build(app_name="charlie", priority=1)
        t2 = QueuedTaskFactory.build(app_name="alpha", priority=2)
        t3 = QueuedTaskFactory.build(app_name="charlie", priority=1)
        data = [t1, t2, t3]
        qs = ListAsQuerySet(data, model=QueuedTask)

        # when
        result = qs.values_list("app_name", flat=True)

        # then
        self.assertListEqual(result, ["charlie", "alpha", "charlie"])

    def test_should_return_flat_and_distinct(self):
        # given
        t1 = QueuedTaskFactory.build(app_name="charlie", priority=1)
        t2 = QueuedTaskFactory.build(app_name="alpha", priority=2)
        t3 = QueuedTaskFactory.build(app_name="charlie", priority=1)
        data = [t1, t2, t3]
        qs = ListAsQuerySet(data, model=QueuedTask)

        # when
        result = qs.distinct().values_list("app_name", flat=True)

        # then
        self.assertListEqual(result, ["charlie", "alpha"])

    def test_should_raise_error_when_trying_flat_with_multiple_fields(self):
        # given
        t1 = QueuedTaskFactory.build(name="charlie", priority=1)
        t2 = QueuedTaskFactory.build(name="alpha", priority=2)
        data = [t1, t2]
        qs = ListAsQuerySet(data, model=QueuedTask)

        # when/then
        with self.assertRaises(TypeError):
            qs.values_list("name", "priority", flat=True)
